import asyncio
from datetime import datetime, timedelta
from functools import lru_cache
import inspect
import logging
import pkgutil
import random
from typing import List, Type
from importlib import import_module

from loguru import logger

from ..utils import next_random_datetime
from ..log import formatter
from . import __name__ as __product__
from .link import Link
from .log import TelegramStream
from .tele import ClientsSession
from .bots.base import BaseBotCheckin

logger = logger.bind(scheme="telegram")


def get_spec(type: str):
    """服务模块路径解析."""
    if type == "checkiner":
        sub = "bots"
        suffix = "checkin"
    elif type == "monitor":
        sub = "monitor"
        suffix = "monitor"
    elif type == "messager":
        sub = "messager"
        suffix = "messager"
    else:
        raise ValueError(f"{type} is not a valid service.")
    return sub, suffix


@lru_cache
def get_names(type: str, allow_ignore=False) -> List[str]:
    """列出服务中所有可用站点."""
    sub, _ = get_spec(type)
    results = []
    typemodule = import_module(f"{__product__}.{sub}")
    for _, mn, _ in pkgutil.iter_modules(typemodule.__path__):
        module = import_module(f"{__product__}.{sub}.{mn}")
        if not allow_ignore:
            if not getattr(module, "__ignore__", False):
                results.append(mn)
        else:
            if (not mn == "base") and (not mn == "test"):
                results.append(mn)
    return results


def get_cls(type: str, names: List[str] = None) -> List[Type]:
    """获得服务特定站点的所有类."""
    sub, suffix = get_spec(type)
    if names == None:
        names = get_names(type)
    results = []
    for name in names:
        try:
            module = import_module(f"{__product__}.{sub}.{name.lower()}")
            for cn, cls in inspect.getmembers(module, inspect.isclass):
                if (name.lower().replace("_", "") + suffix).lower() == cn.lower():
                    results.append(cls)
        except ImportError:
            all_names = get_names(type)
            logger.warning(f'您配置的 "{type}" 不支持站点 "{name}", 请从以下站点中选择:')
            logger.warning(", ".join(all_names))
    return results


def extract(clss: List[Type]) -> List[Type]:
    """对于嵌套类, 展开所有子类."""
    extracted = []
    for cls in clss:
        ncs = [c for c in cls.__dict__.values() if inspect.isclass(c)]
        if ncs:
            extracted.extend(ncs)
        else:
            extracted.append(cls)
    return extracted


async def _checkin_task(checkiner: BaseBotCheckin, sem, wait=0):
    """签到器壳, 用于随机等待开始."""
    if wait > 0:
        checkiner.log.debug(f"随机启动等待: 将等待 {wait} 分钟以启动.")
    await asyncio.sleep(wait * 60)
    async with sem:
        return await checkiner._start()


async def _gather_task(tasks, username):
    return username, await asyncio.gather(*tasks)


async def checkiner(config: dict, instant=False):
    """签到器入口函数."""
    logger.debug("正在启动每日签到模块, 请等待登录.")
    async with ClientsSession.from_config(config) as clients:
        coros = []
        async for tg in clients:
            log = logger.bind(scheme="telechecker", username=tg.me.name)
            if not await Link(tg).auth("checkiner"):
                log.error(f"功能初始化失败: 权限校验不通过.")
                continue
            sem = asyncio.Semaphore(int(config.get("concurrent", 1)))
            clses = extract(get_cls("checkiner", names=config.get("service", {}).get("checkiner", None)))
            checkiners = [
                cls(
                    tg,
                    retries=config.get("retries", 4),
                    timeout=config.get("timeout", 120),
                    nofail=config.get("nofail", True),
                    basedir=config.get("basedir", None),
                    proxy=config.get("proxy", None),
                    config=config.get("checkiner", {}).get(cls.__module__.rsplit(".", 1)[-1], {}),
                )
                for cls in clses
            ]
            tasks = []
            names = []
            for c in checkiners:
                names.append(c.name)
                wait = 0 if instant else random.uniform(0, int(config.get("random", 15)))
                task = asyncio.create_task(_checkin_task(c, sem, wait))
                tasks.append(task)
            coros.append(asyncio.ensure_future(_gather_task(tasks, username=tg.me.name)))
            if names:
                log.debug(f'已启用签到器: {", ".join(names)}')
        while coros:
            done, coros = await asyncio.wait(coros, return_when=asyncio.FIRST_COMPLETED)
            for t in done:
                try:
                    username, results = await t
                except asyncio.CancelledError:
                    continue
                log = logger.bind(scheme="telechecker", username=username)
                failed = []
                ignored = []
                successful = []
                for i, c in enumerate(checkiners):
                    if results[i] == False:
                        failed.append(c)
                    elif results[i] is None:
                        ignored.append(c)
                    else:
                        successful.append(c)
                spec = f"共{len(checkiners)}个"
                if successful:
                    spec += f", {len(successful)}成功"
                if failed:
                    spec += f", {len(failed)}失败"
                if ignored:
                    spec += f", {len(ignored)}跳过"
                if failed:
                    log.error(f"签到失败 ({spec}): {', '.join([f.name for f in failed])}")
                else:
                    log.bind(notify=True).info(f"签到成功 ({spec}).")


async def checkiner_schedule(config: dict, start_time=None, end_time=None, days: int = 0, instant=False):
    """签到器计划任务."""
    while True:
        dt = next_random_datetime(start_time, end_time, interval_days=days)
        logger.bind(scheme="telechecker").info(f"下一次签到将在 {dt.strftime('%m-%d %H:%M %p')} 进行.")
        await asyncio.sleep((dt - datetime.now()).total_seconds())
        await checkiner(config, instant=instant)


async def monitorer(config: dict):
    """监控器入口函数."""
    logger.debug("正在启动消息监控模块.")
    jobs = []
    async with ClientsSession.from_config(config, monitor=True) as clients:
        async for tg in clients:
            log = logger.bind(scheme="telemonitor", username=tg.me.name)
            if not await Link(tg).auth("monitorer"):
                log.error(f"功能初始化失败: 权限校验不通过.")
                continue
            clses = extract(get_cls("monitor", names=config.get("service", {}).get("monitor", None)))
            names = []
            for cls in clses:
                cls_config = config.get("monitor", {}).get(cls.__module__.rsplit(".", 1)[-1], {})
                jobs.append(
                    asyncio.create_task(
                        cls(
                            tg,
                            nofail=config.get("nofail", True),
                            basedir=config.get("basedir", None),
                            proxy=config.get("proxy", None),
                            config=cls_config,
                        )._start()
                    )
                )
                names.append(cls.name)
            if names:
                log.debug(f'已启用监控器: {", ".join(names)}')
        await asyncio.gather(*jobs)


async def messager(config: dict):
    """自动回复器入口函数."""
    logger.debug("正在启动自动水群模块.")
    messagers = []
    async with ClientsSession.from_config(config, send=True) as clients:
        async for tg in clients:
            log = logger.bind(scheme="telemessager", username=tg.me.name)
            if not await Link(tg).auth("messager"):
                log.error(f"功能初始化失败: 权限校验不通过.")
                continue
            clses = extract(get_cls("messager", names=config.get("service", {}).get("messager", None)))
            for cls in clses:
                cls_config = config.get("messager", {}).get(cls.__module__.rsplit(".", 1)[-1], {})
                messagers.append(
                    cls(
                        {"api_id": tg.api_id, "api_hash": tg.api_hash, "phone": tg.phone_number},
                        username=tg.me.name,
                        nofail=config.get("nofail", True),
                        proxy=config.get("proxy", None),
                        basedir=config.get("basedir", None),
                        config=cls_config,
                    )
                )
    await asyncio.gather(*[m.start() for m in messagers])


async def start_notifier(config: dict):
    """消息通知初始化函数."""

    def _filter(record):
        notify = record.get("extra", {}).get("notify", None)
        if notify or record["level"].no == logging.ERROR:
            return True
        else:
            return False

    def _formatter(record):
        notify = record.get("extra", {}).get("notify", False)
        format = formatter(record)
        if notify and notify != True:
            format = format.replace("{message}", "{extra[notify]}")
        return "{level}#" + format

    accounts = config.get("telegram", [])
    notifier = config.get("notifier", None)
    if notifier:
        try:
            if notifier == True:
                notifier = accounts[0]
            elif isinstance(notifier, int):
                notifier = accounts[notifier + 1]
            elif isinstance(notifier, str):
                for a in accounts:
                    if a["phone"] == notifier:
                        notifier = a
                        break
            else:
                notifier = None
        except IndexError:
            notifier = None
    if notifier:
        logger.info(f'计划任务的关键消息将通过 Embykeeper Bot 发送至 "{notifier["phone"]}" 账号.')
        logger.add(
            TelegramStream(
                account=notifier, proxy=config.get("proxy", None), basedir=config.get("basedir", None)
            ),
            format=_formatter,
            filter=_filter,
        )



import asyncio
import random
from typing import Iterable, Union
from datetime import datetime, time, timedelta, timezone, tzinfo
import uuid

from aiohttp import ClientError, ClientConnectionError
from loguru import logger
from embypy.objects import Episode, Movie

from ..utils import show_exception, next_random_datetime, truncate_str
from ..var import debug
from .emby import Emby, Connector, EmbyObject

logger = logger.bind(scheme="embywatcher")


class PlayError(Exception):
    pass


def is_ok(co):
    """判定返回来自 emby 的响应为成功."""
    if isinstance(co, tuple):
        co, *_ = co
    if 200 <= co < 300:
        return True


async def get_random_media(emby: Emby):
    """获取随机视频."""
    while True:
        items = await emby.get_items(["Movie", "Episode"], limit=10, sort="Random", ascending=False)
        i: Union[Movie, Episode]
        for i in items:
            yield i


async def set_played(obj: EmbyObject):
    """设定已播放."""
    c: Connector = obj.connector
    return is_ok(await c.post(f"/Users/{{UserId}}/PlayedItems/{obj.id}"))


async def hide_from_resume(obj: EmbyObject):
    """从首页的"继续收看"部分隐藏."""
    c: Connector = obj.connector
    try:
        return is_ok(await c.post(f"/Users/{{UserId}}/Items/{obj.id}/HideFromResume", Hide=True))
    except RuntimeError:
        return False


def get_last_played(obj: EmbyObject):
    """获取上次播放时间."""
    last_played = obj.object_dict.get("UserData", {}).get("LastPlayedDate", None)
    return datetime.fromisoformat(last_played[:-2]) if last_played else None


async def play(obj: EmbyObject, time: float = 10):
    """模拟播放视频."""
    c: Connector = obj.connector

    # 检查
    totalticks = obj.object_dict.get("RunTimeTicks")
    if not totalticks:
        raise PlayError("无法获取视频长度")
    if time:
        if totalticks < time * 10000000:
            raise PlayError("视频长度低于观看进度所需")
    else:
        time = totalticks / 10000000

    # 获取播放源
    resp = await c.postJson(
        f"/Items/{obj.id}/PlaybackInfo",
        UserID=c.userid,
        StartTimeTicks=0,
        IsPlayback=True,
        AutoOpenLiveStream=True,
    )
    play_session_id = resp.get("PlaySessionId", "")
    if "MediaSources" in resp:
        media_source_id = resp["MediaSources"][0]["Id"]
    else:
        media_source_id = obj.id

    # 模拟播放
    playing_info = lambda tick: {
        "VolumeLevel": 100,
        "CanSeek": True,
        "BufferedRanges": [{"start": 0, "end": tick}] if tick else [],
        "IsPaused": False,
        "ItemId": obj.id,
        "MediaSourceId": media_source_id,
        "PlayMethod": "DirectStream",
        "PlaySessionId": play_session_id,
        "PlaybackStartTimeTicks": datetime.now().timestamp() * 10000000,
        "PlaylistIndex": 0,
        "PlaylistLength": 1,
        "PositionTicks": tick,
        "RepeatMode": "RepeatNone",
    }

    if not is_ok(await c.post("Sessions/Playing", MediaSourceId=media_source_id, data=playing_info(0.1))):
        raise PlayError("无法开始播放")

    t = time
    last_report_t = t
    while t > 0:
        if last_report_t and last_report_t - t > (5 if debug else 30):
            logger.info(f'正在播放: "{truncate_str(obj.name, 10)}" (还剩 {t:.0f} 秒).')
            last_report_t = t
        st = random.uniform(3, 5)
        await asyncio.sleep(st)
        t -= st
        tick = int((time - t) * 10000000)
        payload = playing_info(tick)
        try:
            resp = await asyncio.wait_for(
                c.post("/Sessions/Playing/Progress", data=payload, EventName="timeupdate"), 10
            )
        except (
            ClientError,
            ConnectionError,
            TimeoutError,
            asyncio.TimeoutError,
            asyncio.IncompleteReadError,
        ) as e:
            if isinstance(e, asyncio.IncompleteReadError):
                await c._reset_session()
            logger.debug(f"播放状态设定错误: {e}")
        else:
            if not is_ok(resp):
                logger.debug(f"播放状态设定错误: {resp}")

    if not is_ok(await c.post("/Sessions/Playing/Stopped", data=playing_info(time * 10000000))):
        raise PlayError("无法停止播放")

    return True


async def login(config, continuous=False):
    """登录账号."""
    for a in config.get("emby", ()):
        if not continuous == a.get("continuous", False):
            continue
        logger.info(f'登录账号: {a["username"]} @ {a["url"]}')
        emby = Emby(
            url=a["url"],
            username=a["username"],
            password=a["password"],
            jellyfin=a.get("jellyfin", False),
            proxy=config.get("proxy", None),
            device_id=str(uuid.uuid4()).upper(),
        )
        try:
            info = await emby.info()
        except (ConnectionError, RuntimeError, ClientConnectionError) as e:
            logger.error(f'Emby ({a["url"]}) 连接错误, 请重新检查配置: {e}')
            continue
        if info:
            loggeruser = logger.bind(server=info["ServerName"], username=a["username"])
            loggeruser.info(f'成功登录 ({"Jellyfin" if a.get("jellyfin", False) else "Emby"} {info["Version"]}).')
            yield emby, a.get("time", None if continuous else [120, 240]), loggeruser
        else:
            logger.error(f'Emby ({a["url"]}) 无法获取元信息而跳过, 请重新检查配置.')
            continue


async def watch(emby, time, logger, retries=5):
    """
    主执行函数 - 观看一个视频.
    参数:
        emby: Emby 客户端
        time: 模拟播放时间
        progress: 播放后设定的观看进度
        logger: 日志器
        retries: 最大重试次数
    """
    retry = 0
    while True:
        try:
            async for obj in get_random_media(emby):
                if isinstance(time, Iterable):
                    t = random.uniform(*time)
                else:
                    t = time
                logger.info(f'开始尝试播放 "{truncate_str(obj.name, 10)}" ({t:.0f} 秒).')
                while True:
                    try:
                        await play(obj, t)

                        obj = await obj.update("UserData")

                        if obj.play_count < 1:
                            raise PlayError("尝试播放后播放数低于1")
                        last_played = get_last_played(obj)

                        if not last_played:
                            raise PlayError("尝试播放后无记录")
                        last_played = (
                            last_played.replace(tzinfo=timezone.utc)
                            .astimezone(tz=None)
                            .strftime("%m-%d %H:%M")
                        )

                        prompt = (
                            f"[yellow]成功播放视频[/], " + f"当前该视频播放 {obj.play_count} 次, " + f"上次播放于 {last_played}."
                        )
                        logger.bind(notify="成功保活.").info(prompt)
                        return True
                    except (ClientError, OSError, asyncio.IncompleteReadError) as e:
                        retry += 1
                        if retry > retries:
                            logger.warning(f"超过最大重试次数, 保活失败: {e}.")
                            return False
                        else:
                            if isinstance(e, asyncio.IncompleteReadError):
                                await emby.connector._reset_session()
                            rt = random.uniform(30, 60)
                            logger.info(f"连接失败, 等待 {rt:.0f} 秒后重试: {e}.")
                            await asyncio.sleep(rt)
                    except PlayError as e:
                        retry += 1
                        if retry > retries:
                            logger.warning(f"超过最大重试次数, 保活失败: {e}.")
                            return False
                        else:
                            rt = random.uniform(30, 60)
                            logger.info(f"发生错误, 等待 {rt:.0f} 秒后重试其他视频: {e}.")
                            await asyncio.sleep(rt)
                        break
                    finally:
                        try:
                            if not await asyncio.shield(asyncio.wait_for(hide_from_resume(obj), 5)):
                                logger.debug(f"未能成功从最近播放中隐藏视频.")
                        except asyncio.TimeoutError:
                            logger.debug(f"从最近播放中隐藏视频超时.")
                        else:
                            logger.info(f"已从最近播放中隐藏该视频.")
            else:
                logger.warning(f"由于没有成功播放视频, 保活失败, 请重新检查配置.")
                return False
        except (ClientError, OSError, asyncio.IncompleteReadError) as e:
            retry += 1
            if retry > retries:
                logger.warning(f"超过最大重试次数, 保活失败: {e}.")
                return False
            else:
                if isinstance(e, asyncio.IncompleteReadError):
                    await emby.connector._reset_session()
                rt = random.uniform(30, 60)
                logger.info(f"连接失败, 等待 {rt:.0f} 秒后重试其他视频: {e}.")
                await asyncio.sleep(rt)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logger.warning(f"发生错误, 保活失败.")
            show_exception(e, regular=False)
            return False


async def watch_continuous(emby: Emby, logger):
    """
    主执行函数 - 持续观看.

    参数:
        emby: Emby 客户端
        logger: 日志器
    """
    while True:
        try:
            async for obj in get_random_media(emby):
                totalticks = obj.object_dict.get("RunTimeTicks")
                if not totalticks:
                    raise PlayError("无法获取视频长度")
                logger.info(f'开始尝试播放 "{truncate_str(obj.name, 10)}" (长度 {totalticks / 10000000:.0f} 秒).')
                try:
                    await play(obj, 0)
                except PlayError as e:
                    rt = random.uniform(30, 60)
                    logger.info(f"发生错误, 等待 {rt:.0f} 秒后重试: {e}.")
                    await asyncio.sleep(rt)
                    continue
                finally:
                    try:
                        if not await asyncio.shield(asyncio.wait_for(hide_from_resume(obj), 2)):
                            logger.debug(f"未能成功从最近播放中隐藏视频.")
                    except asyncio.TimeoutError:
                        logger.debug(f"从最近播放中隐藏视频超时.")
        except (ClientError, OSError) as e:
            if isinstance(e, asyncio.IncompleteReadError):
                await emby.connector._reset_session()
            rt = random.uniform(30, 60)
            logger.info(f"连接失败, 等待 {rt:.0f} 秒后重试: {e}.")
            await asyncio.sleep(rt)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logger.warning(f"发生错误, 停止持续播放.")
            show_exception(e, regular=False)
            return False


async def watcher(config: dict):
    """入口函数 - 观看一个视频."""

    async def wrapper(emby, time, logger):
        try:
            if isinstance(time, Iterable):
                tm = max(time) * 2
            else:
                tm = time * 2
            return await asyncio.wait_for(watch(emby, time, logger), max(tm, 180))
        except asyncio.TimeoutError:
            logger.warning(f"一定时间内未完成播放, 保活失败.")
            return False

    tasks = []
    async for emby, time, loggeruser in login(config):
        tasks.append(wrapper(emby, time, loggeruser))
    results = await asyncio.gather(*tasks)
    fails = len(tasks) - sum(results)
    if fails:
        logger.error(f"保活失败 ({fails}/{len(tasks)}).")


async def watcher_schedule(config: dict, start_time=time(11, 0), end_time=time(23, 0), days: int = 7):
    """计划任务 - 观看一个视频."""
    while True:
        dt = next_random_datetime(start_time, end_time, interval_days=days)
        logger.bind(scheme="embywatcher").info(f"下一次保活将在 {dt.strftime('%m-%d %H:%M %p')} 进行.")
        await asyncio.sleep((dt - datetime.now()).total_seconds())
        await watcher(config)


async def watcher_continuous(config: dict):
    """入口函数 - 持续观看."""

    async def wrapper(emby, time, logger):
        if time:
            if isinstance(time, Iterable):
                time = random.uniform(*time)
            logger.info(f"即将连续播放视频, 持续 {time:.0f} 秒.")
        else:
            logger.info(f"即将无限连续播放视频.")
        try:
            await asyncio.wait_for(watch_continuous(emby, logger), time)
        except asyncio.TimeoutError:
            return True
        else:
            return False

    tasks = []
    async for emby, time, logger in login(config, continuous=True):
        tasks.append(wrapper(emby, time, logger))
    return await asyncio.gather(*tasks)


async def watcher_continuous_schedule(
    config: dict, start_time=time(11, 0), end_time=time(23, 0), days: int = 1
):
    """计划任务 - 持续观看."""

    dt = datetime.now() + 10
    t = asyncio.create_task(watcher_continuous(config))
    while True:
        dt = next_random_datetime(start_time, end_time, interval_days=days)
        logger.bind(scheme="embywatcher").info(f"持续观看结束后, 将在 {dt.strftime('%m-%d %H:%M %p')} 再次开始.")
        await asyncio.sleep((dt - datetime.now()).total_seconds())
        if not t.done():
            t.cancel()
        await watcher(config)



from pyrogram.types import Message

from .base import AnswerBotCheckin

__ignore__ = True


class JMSIPTVCheckin(AnswerBotCheckin):
    name = "卷毛鼠IPTV"
    bot_username = "JMSIPTV_bot"
    bot_captcha_len = 5

    async def on_captcha(self, message: Message, captcha: str):
        await super().on_captcha(message, captcha.upper())



import os
import platform
import sys
import time
from msvcrt import getch
from pathlib import Path
from subprocess import Popen

from appdirs import user_data_dir
from rich import box
from rich.align import Align
from rich.panel import Panel
from rich.table import Table

from . import __name__ as __product__
from . import var
from .cli import app as cli
from .settings import write_faked_config


def generate_config(config: Path):
    if config.exists():
        return False
    os.makedirs(config.parent)
    write_faked_config(config, quiet=True)
    message = Table.grid(padding=2)
    message.add_column()
    urls = Table.grid(padding=2)
    urls.add_column(style="green", justify="right")
    urls.add_column(no_wrap=True)
    urls.add_row("配置教程", "https://github.com/embykeeper/embykeeper/wiki/配置文件")
    message.add_row("您需要在即将打开的文件中进行账户配置, 您可以根据注释提示进行配置.")
    message.add_row("您也可以参考 Wiki, 链接如下:")
    message.add_row(urls)
    message.add_row(Align("请按任意键以打开模板配置文件", align="center", style="bold blink underline"))
    var.console.print(
        Panel.fit(
            message,
            box=box.ROUNDED,
            padding=(1, 2),
            title="[b red]欢迎您使用 Embykeeper!",
            border_style="bright_blue",
        ),
        justify="center",
    )

    _ = getch()

    p = Popen(["start", config], shell=True)

    if not p:
        var.console.print(f"配置文件打开失败, 请按任意键退出", justify="center")
        _ = getch()
        sys.exit(1)

    var.console.print(f"等待您编辑配置文件, 请保存关闭编辑器窗口以继续 ...", justify="center")
    p.wait()
    var.console.print(f"请确认您配置完成, 并按任意键以继续启动 {__product__.capitalize()}...", justify="center")
    var.console.print(f"配置完成, 即将启动 {__product__.capitalize()} ...", justify="center")


def main():
    config = Path(user_data_dir(__product__)) / "config.toml"
    generate_config(config)
    os.system("cls")
    var.console.rule("Embykeeper")
    cli(["-W"])


if __name__ == "__main__":
    if not platform.system() == "Windows":
        raise RuntimeError("this entrypoint is only for windows")
    main()



from __future__ import annotations
from collections import OrderedDict
from contextlib import asynccontextmanager

from datetime import datetime
import asyncio
import inspect
from pathlib import Path
import pickle
import random
import sys
from typing import AsyncGenerator, Optional, Union
import uuid

from rich.prompt import Prompt
from appdirs import user_data_dir
from loguru import logger
import pyrogram
from pyrogram import raw, types, utils, filters, dispatcher
from pyrogram.enums import SentCodeType
from pyrogram.errors import (
    BadRequest,
    RPCError,
    ApiIdPublishedFlood,
    Unauthorized,
    SessionPasswordNeeded,
    CodeInvalid,
    PhoneCodeInvalid,
)
from pyrogram.handlers import MessageHandler, RawUpdateHandler, DisconnectHandler
from pyrogram.handlers.handler import Handler
from aiocache import Cache

from .. import var, __name__, __version__
from ..utils import async_partial, show_exception, to_iterable

logger = logger.bind(scheme="telegram")

_id = b"\x80\x04\x95\x15\x00\x00\x00\x00\x00\x00\x00]\x94(K2K2K9K7K9K6K4K8e."
_hash = b"\x80\x04\x95E\x00\x00\x00\x00\x00\x00\x00]\x94(K7K8KeKeKfKcKfKbK9K8K9KeK1K1K0KcK0KdK3K0K7K8K3K8K5KfK9K9K7KaKeKee."
_decode = lambda x: "".join(map(chr, to_iterable(pickle.loads(x))))

# 密钥信息
API_KEY = {
    "_": {"api_id": _decode(_id), "api_hash": _decode(_hash)}
    # "nicegram": {"api_id": "94575", "api_hash": "a3406de8d171bb422bb6ddf3bbd800e2"},
    # "tgx-android": {"api_id": "21724", "api_hash": "3e0cb5efcd52300aec5994fdfc5bdc16"},
    # "tg-react": {"api_id": "414121", "api_hash": "db09ccfc2a65e1b14a937be15bdb5d4b"},
}


def _name(self: Union[types.User, types.Chat]):
    return " ".join([n for n in (self.first_name, self.last_name) if n])


def _chat_name(self: types.Chat):
    if self.title:
        return self.title
    else:
        return _name(self)


setattr(types.User, "name", property(_name))
setattr(types.Chat, "name", property(_chat_name))


class Dispatcher(dispatcher.Dispatcher):
    def __init__(self, client: Client):
        super().__init__(client)
        self.mutex = asyncio.Lock()

    async def start(self):
        logger.debug("Telegram 更新分配器启动.")
        if not self.client.no_updates:
            self.handler_worker_tasks = [
                self.loop.create_task(self.handler_worker()) for _ in range(self.client.workers)
            ]

    def add_handler(self, handler, group: int):
        async def fn():
            async with self.mutex:
                if group not in self.groups:
                    self.groups[group] = []
                    self.groups = OrderedDict(sorted(self.groups.items()))
                self.groups[group].append(handler)
                # logger.debug(f"增加了 Telegram 更新处理器: {handler.__class__.__name__}.")

        return self.loop.create_task(fn())

    def remove_handler(self, handler, group: int):
        async def fn():
            async with self.mutex:
                if group not in self.groups:
                    raise ValueError(f"Group {group} does not exist. Handler was not removed.")
                self.groups[group].remove(handler)
                # logger.debug(f"移除了 Telegram 更新处理器: {handler.__class__.__name__}.")

        return self.loop.create_task(fn())

    async def handler_worker(self):
        while True:
            packet = await self.updates_queue.get()

            if packet is None:
                break

            try:
                update, users, chats = packet
                parser = self.update_parsers.get(type(update), None)

                parsed_update, handler_type = (
                    await parser(update, users, chats) if parser is not None else (None, type(None))
                )

                async with self.mutex:
                    groups = {i: g[:] for i, g in self.groups.items()}

                for group in groups.values():
                    for handler in group:
                        args = None

                        if isinstance(handler, handler_type):
                            try:
                                if await handler.check(self.client, parsed_update):
                                    args = (parsed_update,)
                            except Exception as e:
                                logger.warning(f"Telegram 错误: {e}")
                                continue

                        elif isinstance(handler, RawUpdateHandler):
                            args = (update, users, chats)

                        if args is None:
                            continue

                        try:
                            if inspect.iscoroutinefunction(handler.callback):
                                await handler.callback(self.client, *args)
                            else:
                                await self.loop.run_in_executor(
                                    self.client.executor, handler.callback, self.client, *args
                                )
                        except pyrogram.StopPropagation:
                            raise
                        except pyrogram.ContinuePropagation:
                            continue
                        except Exception as e:
                            logger.error(f"更新回调函数内发生错误.")
                            show_exception(e, regular=False)
                        break
            except pyrogram.StopPropagation:
                pass
            except Exception as e:
                logger.error("更新控制器错误.")
                show_exception(e, regular=False)


class Client(pyrogram.Client):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.cache = Cache()
        self.dispatcher = Dispatcher(self)

    async def authorize(self):
        if self.bot_token:
            return await self.sign_in_bot(self.bot_token)
        retry = False
        while True:
            try:
                sent_code = await self.send_code(self.phone_number)
                code_target = {
                    SentCodeType.APP: " Telegram 客户端",
                    SentCodeType.SMS: "短信",
                    SentCodeType.CALL: "来电",
                    SentCodeType.FLASH_CALL: "闪存呼叫",
                    SentCodeType.FRAGMENT_SMS: " Fragment 短信",
                    SentCodeType.EMAIL_CODE: "邮件",
                }
                if not self.phone_code:
                    if retry:
                        msg = f'验证码错误, 请重新输入 "{self.phone_number}" 的登录验证码 (按回车确认)'
                    else:
                        msg = f'请从{code_target[sent_code.type]}接收 "{self.phone_number}" 的登录验证码 (按回车确认)'
                    self.phone_code = Prompt.ask(" " * 23 + msg, console=var.console)
                signed_in = await self.sign_in(self.phone_number, sent_code.phone_code_hash, self.phone_code)
            except (CodeInvalid, PhoneCodeInvalid):
                self.phone_code = None
                retry = True
            except SessionPasswordNeeded:
                retry = False
                while True:
                    if not self.password:
                        if retry:
                            msg = f'密码错误, 请重新输入 "{self.phone_number}" 的两步验证密码 (不显示, 按回车确认)'
                        else:
                            msg = f'需要输入 "{self.phone_number}" 的两步验证密码 (不显示, 按回车确认)'
                        self.password = Prompt.ask(" " * 23 + msg, password=True, console=var.console)
                    try:
                        return await self.check_password(self.password)
                    except BadRequest:
                        self.password = None
                        retry = True
            else:
                break
        if isinstance(signed_in, types.User):
            return signed_in
        else:
            raise BadRequest("该账户尚未注册")

    def add_handler(self, handler: Handler, group: int = 0):
        if isinstance(handler, DisconnectHandler):
            self.disconnect_handler = handler.callback

            async def dummy():
                pass

            return asyncio.ensure_future(dummy())
        else:
            return self.dispatcher.add_handler(handler, group)

    def remove_handler(self, handler: Handler, group: int = 0):
        if isinstance(handler, DisconnectHandler):
            self.disconnect_handler = None

            async def dummy():
                pass

            return asyncio.ensure_future(dummy())
        else:
            return self.dispatcher.remove_handler(handler, group)

    async def get_dialogs(
        self, limit: int = 0, exclude_pinned=None, folder_id=None
    ) -> Optional[AsyncGenerator["types.Dialog", None]]:
        cache_id = f"dialogs_{self.phone_number}_{folder_id}_{1 if exclude_pinned else 0}"
        (offset_id, offset_date, offset_peer), cache = await self.cache.get(
            cache_id, ((0, 0, raw.types.InputPeerEmpty()), [])
        )

        current = 0
        total = limit or (1 << 31) - 1
        limit = min(100, total)

        for c in cache:
            yield c
            current += 1
            if current >= total:
                return

        while True:
            r = await self.invoke(
                raw.functions.messages.GetDialogs(
                    offset_date=offset_date,
                    offset_id=offset_id,
                    offset_peer=offset_peer,
                    limit=limit,
                    hash=0,
                    exclude_pinned=exclude_pinned,
                    folder_id=folder_id,
                ),
                sleep_threshold=60,
            )

            users = {i.id: i for i in r.users}
            chats = {i.id: i for i in r.chats}

            messages = {}

            for message in r.messages:
                if isinstance(message, raw.types.MessageEmpty):
                    continue

                chat_id = utils.get_peer_id(message.peer_id)
                messages[chat_id] = await types.Message._parse(self, message, users, chats)

            dialogs = []

            for dialog in r.dialogs:
                if not isinstance(dialog, raw.types.Dialog):
                    continue

                dialogs.append(types.Dialog._parse(self, dialog, messages, users, chats))

            if not dialogs:
                return

            last = dialogs[-1]

            offset_id = last.top_message.id
            offset_date = utils.datetime_to_timestamp(last.top_message.date)
            offset_peer = await self.resolve_peer(last.chat.id)

            _, cache = await self.cache.get(cache_id, ((0, 0, raw.types.InputPeerEmpty()), []))
            await self.cache.set(cache_id, ((offset_id, offset_date, offset_peer), cache + dialogs), ttl=120)
            for dialog in dialogs:
                yield dialog

                current += 1

                if current >= total:
                    return

    @asynccontextmanager
    async def catch_reply(self, chat_id: Union[int, str], outgoing=False):
        async def handler_func(client, message, future: asyncio.Future):
            future.set_result(message)

        future = asyncio.Future()
        filter = filters.chat(chat_id)
        if not outgoing:
            filter = filter & (~filters.outgoing)
        handler = MessageHandler(async_partial(handler_func, future=future), filter)
        await self.add_handler(handler, group=0)
        try:
            yield future
        finally:
            self.remove_handler(handler, group=0)

    async def wait_reply(
        self, chat_id: Union[int, str], send: str = None, timeout: float = 10, outgoing=False
    ):
        async with self.catch_reply(chat_id=chat_id, outgoing=outgoing) as f:
            if send:
                await self.send_message(chat_id, send)
            msg: types.Message = await asyncio.wait_for(f, timeout)
            return msg

    async def mute_chat(self, chat_id: Union[int, str], until: Union[int, datetime]):
        if isinstance(until, datetime):
            until = until.timestamp()

        return await self.invoke(
            raw.functions.account.UpdateNotifySettings(
                peer=raw.types.InputNotifyPeer(peer=await self.resolve_peer(chat_id)),
                settings=raw.types.InputPeerNotifySettings(
                    show_previews=False,
                    mute_until=int(until),
                ),
            )
        )


class ClientsSession:
    pool = {}
    lock = asyncio.Lock()
    watch = None

    @classmethod
    def from_config(cls, config, **kw):
        accounts = config.get("telegram", [])
        for k, v in kw.items():
            accounts = [a for a in accounts if a.get(k, None) in to_iterable(v)]
        return cls(
            accounts=accounts,
            proxy=config.get("proxy", None),
            basedir=config.get("basedir", None),
        )

    @classmethod
    async def watchdog(cls, timeout=120):
        logger.debug("Telegram 账号池看门狗启动.")
        try:
            counter = {}
            while True:
                await asyncio.sleep(10)
                for p in list(cls.pool):
                    try:
                        if cls.pool[p][1] <= 0:
                            if p in counter:
                                counter[p] += 1
                                if counter[p] >= timeout / 10:
                                    counter[p] = 0
                                    await cls.clean(p)
                            else:
                                counter[p] = 1
                        else:
                            counter.pop(p, None)
                    except (TypeError, KeyError):
                        pass
        except asyncio.CancelledError:
            await cls.shutdown()

    @classmethod
    async def clean(cls, phone):
        async with cls.lock:
            entry = cls.pool.get(phone, None)
            if not entry:
                return
            try:
                client, ref = entry
            except TypeError:
                return
            if not ref:
                logger.debug(f'登出账号 "{client.phone_number}".')
                await client.stop()
                cls.pool.pop(phone, None)

    @classmethod
    async def clean_all(cls):
        for phone in list(cls.pool):
            await cls.clean(phone)

    @classmethod
    async def shutdown(cls):
        print("\r正在停止...\r", end="", flush=True, file=sys.stderr)
        for v in cls.pool.values():
            if isinstance(v, asyncio.Task):
                v.cancel()
            else:
                client, ref = v
                client.dispatcher.updates_queue.put_nowait(None)
                for t in client.dispatcher.handler_worker_tasks:
                    try:
                        await t
                    except asyncio.CancelledError:
                        pass
        while len(asyncio.all_tasks()) > 1:
            await asyncio.sleep(0.1)
        print(f"Telegram 账号池停止.\r", end="", file=sys.stderr)
        for v in cls.pool.values():
            if isinstance(v, tuple):
                client: Client = v[0]
                await client.storage.save()
                await client.storage.close()
                logger.debug(f'登出账号 "{client.phone_number}".')

    def __init__(self, accounts, proxy=None, basedir=None, quiet=False):
        self.accounts = accounts
        self.proxy = proxy
        self.basedir = basedir or user_data_dir(__name__)
        self.phones = []
        self.done = asyncio.Queue()
        self.quiet = quiet
        if not self.watch:
            self.__class__.watch = asyncio.create_task(self.watchdog())

    async def login(self, account, proxy):
        try:
            account["phone"] = "".join(account["phone"].split())
            session_file = Path(self.basedir) / f'{account["phone"]}.session'
            session_string_file = Path(self.basedir) / f'{account["phone"]}.login'
            if not self.quiet:
                logger.info(f'登录至账号 "{account["phone"]}".')
            for _ in range(3):
                if account.get("api_id", None) is None or account.get("api_hash", None) is None:
                    account.update(random.choice(list(API_KEY.values())))
                session_string = account.get("session", None)
                if not session_string:
                    if session_string_file.is_file():
                        with open(session_string_file) as f:
                            session_string = f.read().strip()
                in_memory = True
                if not session_string:
                    if session_file.is_file():
                        in_memory = False
                try:
                    client = Client(
                        app_version=__version__,
                        device_model="Server " + uuid.uuid1().hex[16:20].upper(),
                        name=account["phone"],
                        api_id=account["api_id"],
                        api_hash=account["api_hash"],
                        phone_number=account["phone"],
                        session_string=session_string,
                        in_memory=in_memory,
                        proxy=proxy,
                        workdir=self.basedir,
                    )
                    await client.start()
                except ApiIdPublishedFlood:
                    logger.warning(f'登录账号 "{account["phone"]}" 时发生 API key 限制, 将被跳过.')
                    logger.warning(f"请您申请自己的 API, 参考: https://blog.iair.top/2023/10/15/embykeeper-api.")
                    break
                except Unauthorized:
                    try:
                        await client.storage.delete()
                    except:
                        pass
                except KeyError as e:
                    logger.warning(f'登录账号 "{account["phone"]}" 时发生异常, 可能是由于网络错误, 将在 3 秒后重试.')
                    show_exception(e)
                    await asyncio.sleep(3)
                else:
                    break
            else:
                logger.error(f'登录账号 "{account["phone"]}" 失败次数超限, 将被跳过.')
        except asyncio.CancelledError:
            raise
        except RPCError as e:
            logger.error(f'登录账号 "{account["phone"]}" 失败 ({e.MESSAGE.format(value=e.value)}), 将被跳过.')
        except Exception as e:
            logger.error(f'登录账号 "{account["phone"]}" 时发生异常, 将被跳过.')
            show_exception(e, regular=False)
        else:
            if not session_string_file.exists():
                with open(session_string_file, "w+", encoding="utf-8") as f:
                    f.write(await client.export_session_string())
            logger.debug(f'登录账号 "{client.phone_number}" 成功.')
            return client

    async def loginer(self, account):
        client = await self.login(account, proxy=self.proxy)
        if isinstance(client, Client):
            async with self.lock:
                phone = account["phone"]
                self.pool[phone] = (client, 1)
                self.phones.append(phone)
                await self.done.put(client)
                logger.debug(f"Telegram 账号池计数增加: {phone} => 1")
        else:
            await self.done.put(None)

    async def __aenter__(self):
        for a in self.accounts:
            phone = a["phone"]
            try:
                await self.lock.acquire()
                if phone in self.pool:
                    if isinstance(self.pool[phone], asyncio.Task):
                        self.lock.release()
                        await self.pool[phone]
                        await self.lock.acquire()
                    if isinstance(self.pool[phone], asyncio.Task):
                        continue
                    client, ref = self.pool[phone]
                    ref += 1
                    self.pool[phone] = (client, ref)
                    self.phones.append(phone)
                    await self.done.put(client)
                    logger.debug(f"Telegram 账号池计数增加: {phone} => {ref}")
                else:
                    self.pool[phone] = asyncio.create_task(self.loginer(a))
            finally:
                try:
                    self.lock.release()
                except RuntimeError:
                    pass
        return self

    def __aiter__(self):
        async def aiter():
            for _ in range(len(self.accounts)):
                client: Client = await self.done.get()
                if client:
                    yield client

        return aiter()

    async def __aexit__(self, type, value, tb):
        async with self.lock:
            for phone in self.phones:
                entry = self.pool.get(phone, None)
                client, ref = entry
                ref -= 1
                self.pool[phone] = (client, ref)
                logger.debug(f"Telegram 账号池计数降低: {phone} => {ref}")



import asyncio
from pyrogram.types import Message

from .base import Monitor

__ignore__ = True


class PoloMonitor(Monitor):
    name = "Polo"
    chat_name = "poloemby"
    chat_keyword = r"普通可用的注册码:\n([\s\S]*)"
    bot_username = "polo_emby_bot"
    notify_create_name = True
    allow_edit = False

    async def on_trigger(self, message: Message, key, reply):
        for code in key.split("\n"):
            await self.client.send_message(self.bot_username, f"/invite {code}")
            await self.client.send_message(self.bot_username, self.unique_name)
            await asyncio.sleep(0.5)
            self.log.bind(notify=True).info(f'已向Bot发送邀请码: "{code}", 请查看.')



import asyncio
import random
import re
from dataclasses import dataclass
from datetime import datetime, time, timedelta
from pathlib import Path
from typing import Iterable, List, Union

import yaml
from dateutil import parser
from loguru import logger
from pyrogram.enums import ChatMemberStatus
from pyrogram.errors import ChatWriteForbidden, RPCError, SlowmodeWait
from schema import Optional, Schema, SchemaError

from ...data import get_data
from ...utils import random_time, show_exception, time_in_range, truncate_str
from ..tele import ClientsSession

__ignore__ = True


@dataclass(eq=False)
class MessageSchedule:
    """定义一个发送规划, 即在特定时间段内某些消息中的一个有一定几率被发送."""

    messages: Iterable[str]
    at: Union[Iterable[Union[str, time]], Union[str, time]] = ("0:00", "23:59")
    days: int = 0
    possibility: float = 1.0
    multiply: int = 1
    only: str = None

    def next_time(self, days=None, trange=None):
        """生成下一个发送时间"""
        days = days if days is not None else self.days
        trange = trange if trange is not None else self.at
        rtime = datetime.combine(datetime.today() + timedelta(days=days), random_time(*trange))
        if rtime < datetime.now():
            rtime += timedelta(days=1)
        return rtime

    def roll(self, days=None, trange=None):
        skip = False
        if random.random() >= self.possibility:
            skip = True
        if self.only:
            today = datetime.today()
            if self.only.startswith("weekday") and today.weekday() > 4:
                skip = True
            if self.only.startswith("weekend") and today.weekday() < 5:
                skip = True
        return MessagePlan(
            message=random.choice(self.messages),
            at=self.next_time(days=days, trange=trange),
            schedule=self,
            skip=skip,
        )


@dataclass(eq=False)
class MessagePlan:
    """定义一个发送计划, 即在某事件发送某个消息."""

    message: str
    at: datetime
    schedule: MessageSchedule
    skip: bool = False


class Messager:
    """自动水群类."""

    name: str = None  # 水群器名称
    chat_name: str = None  # 群聊的名称
    default_messages: List[str] = []  # 默认的话术列表资源名

    site_last_message_time = None
    site_lock = asyncio.Lock()

    def __init__(self, account, username=None, nofail=True, proxy=None, basedir=None, config: dict = None):
        """
        自动水群类.
        参数:
            account: 账号登录信息
            username: 用户名, 仅用于在登录前的日志输出
            nofail: 启用错误处理外壳, 当错误时报错但不退出
            basedir: 文件存储默认位置
            proxy: 代理配置
            config: 当前水群器的特定配置
        """
        self.account = account
        self.nofail = nofail
        self.proxy = proxy
        self.basedir = basedir
        self.config = config

        self.min_interval = timedelta(
            seconds=config.get("min_interval", config.get("interval", 1800))
        )  # 两条消息间的最小间隔时间
        max_interval = config.get("max_interval", None)
        if max_interval:
            self.max_interval = timedelta(seconds=max_interval)  # 两条消息间的最大间隔时间
            if self.min_interval > self.max_interval:
                raise ValueError("最小间隔不应大于最大间隔")
        else:
            self.max_interval = None

        self.log = logger.bind(scheme="telemessager", name=self.name, username=username)
        self.timeline: List[MessagePlan] = []  # 消息计划序列

    def parse_message_yaml(self, file):
        with open(file, "r") as f:
            data = yaml.safe_load(f)
        schema = Schema(
            {
                "messages": [str],
                Optional("at"): [str],
                Optional("days"): int,
                Optional("possibility"): float,
                Optional("only"): str,
            }
        )
        schema.validate(data)
        at = data.get("at", ("9:00", "23:00"))
        assert len(at) == 2
        at = [parser.parse(t).time() for t in at]
        return MessageSchedule(
            messages=data.get("messages"),
            at=at,
            days=data.get("days", 0),
            possibility=data.get("possibility", 1.0),
            only=data.get("only", None),
        )

    def add(self, schedule: MessageSchedule):
        for _ in range(600):
            plan = schedule.roll()
            for p in self.timeline:
                if time_in_range(p.at - self.min_interval, p.at + self.min_interval, plan.at):
                    break
            else:
                if self.max_interval:
                    if self.timeline:
                        for p in self.timeline:
                            if time_in_range(p.at - self.max_interval, p.at + self.max_interval, plan.at):
                                self.timeline.append(plan)
                                return
                    else:
                        now = datetime.now()
                        if time_in_range(now - self.max_interval, now + self.max_interval, plan.at):
                            self.timeline.append(plan)
                            return
                else:
                    self.timeline.append(plan)
                    return
        else:
            plan.skip = True
            self.timeline.append(plan)
            return

    async def get_spec_path(self, spec):
        if not Path(spec).exists():
            return await get_data(self.basedir, spec, proxy=self.proxy, caller=f"{self.name}水群")
        else:
            return spec

    async def get_spec_schedule(self, spec):
        file = await self.get_spec_path(spec)
        if not file:
            self.log.warning(f'话术文件 "{spec}" 无法下载或不存在, 将被跳过.')
            return None
        try:
            return self.parse_message_yaml(file)
        except (OSError, yaml.YAMLError, SchemaError) as e:
            self.log.warning(f'话术文件 "{spec}" 错误, 将被跳过: {e}')
            show_exception(e)
            return None

    async def start(self):
        messages = self.config.get("messages", [])
        if not messages:
            messages = self.default_messages

        schedules = []
        for m in messages:
            match = re.match(r"(.*)\*\s?(\d+)", m)
            if match:
                multiply = int(match.group(2))
                spec = match.group(1).strip()
            else:
                multiply = 1
                spec = m
            schedule = await self.get_spec_schedule(spec)
            if schedule:
                schedule.multiply = multiply
                schedules.append(schedule)

        nmsgs = sum([s.multiply for s in schedules])
        self.log.info(f"共启用 {len(schedules)} 个消息规划, 发送 {nmsgs} 条消息.")
        for s in schedules:
            for _ in range(s.multiply):
                self.add(s)

        if self.timeline:
            last_valid_p = None
            while True:
                valid_p = [p for p in self.timeline if not p.skip]
                self.log.debug(f"时间线上当前有 {len(self.timeline)} 个消息计划, {len(valid_p)} 个有效.")
                # self.log.debug(
                #     "时间序列: " + " ".join([p.at.strftime("%H%M") for p in sorted(valid_p, key=lambda x: x.at)])
                # )
                if valid_p:
                    next_valid_p = min(valid_p, key=lambda x: x.at)
                    if not next_valid_p == last_valid_p:
                        last_valid_p = next_valid_p
                        self.log.info(
                            f"下一次发送将在 [blue]{next_valid_p.at.strftime('%m-%d %H:%M:%S')}[/] 进行: {truncate_str(next_valid_p.message, 20)}."
                        )
                else:
                    self.log.info(f"下一次发送被跳过.")
                next_p = min(self.timeline, key=lambda x: x.at)
                self.log.debug(
                    f"下一次计划任务将在 [blue]{next_p.at.strftime('%m-%d %H:%M:%S')}[/] 进行 ({'跳过' if next_p.skip else '有效'})."
                )
                await asyncio.sleep((next_p.at - datetime.now()).total_seconds())
                if not next_p.skip:
                    await self._send(next_p.message)
                self.timeline.remove(next_p)
                self.add(next_p.schedule)

    async def _send(self, *args, **kw):
        try:
            return await self.send(*args, **kw)
        except OSError as e:
            self.log.info(f'出现错误: "{e}", 忽略.')
            show_exception(e)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            if self.nofail:
                self.log.warning(f"发生错误, 自动水群将停止.")
                show_exception(e, regular=False)
            else:
                raise

    async def prepare_send(self, message):
        return message

    async def send(self, message):
        message = await self.prepare_send(message)
        if not message:
            return
        if self.site_last_message_time:
            need_sec = random.randint(5, 10)
            while self.site_last_message_time > datetime.now() - timedelta(seconds=need_sec):
                await asyncio.sleep(1)
        async with ClientsSession([self.account], proxy=self.proxy, basedir=self.basedir) as clients:
            async for tg in clients:
                chat = await tg.get_chat(self.chat_name)
                self.log.bind(username=tg.me.name).info(f'向聊天 "{chat.name}" 发送: [gray50]{message}[/]')
                try:
                    await tg.send_message(self.chat_name, message)
                except ChatWriteForbidden as e:
                    try:
                        chat = await tg.get_chat(self.chat_name)
                        member = await chat.get_member(tg.me.id)
                        if member.status == ChatMemberStatus.RESTRICTED:
                            if member.until_date:
                                delta = member.until_date - datetime.now()
                                secs = delta.total_seconds()
                                if secs < 2592000:
                                    self.log.warning(f"发送失败: 您已被禁言 {secs} 秒.")
                                else:
                                    self.log.warning(f"发送失败: 您已被永久禁言.")
                            else:
                                self.log.warning(f"发送失败: 您已被禁言.")
                        else:
                            self.log.warning(f"发送失败: 已全员禁言.")
                    except RPCError:
                        self.log.warning(f"发送失败: {e}.")
                except SlowmodeWait as e:
                    self.log.info(f"发送等待: Telegram 限制, 等待 {e.value} 秒.")
                except RPCError as e:
                    self.log.warning(f"发送失败: {e}.")
                except KeyError as e:
                    self.log.warning(f"发送失败, 您可能已被封禁.")
                    show_exception(e)
                else:
                    async with self.site_lock:
                        self.__class__.site_last_message_time = datetime.now()



from .base import Messager

__ignore__ = True


class NakonakoMessager(Messager):
    name = "NakoNako"
    chat_name = "NakoNetwork"
    default_messages = [
        "watery-wl@v1.yaml * 10",
        "goodday-wl@v1.yaml",
        "goodnoon-wl@v1.yaml",
        "goodeve-wl@v1.yaml",
        "goodnight-wl@v1.yaml",
    ]



from pyrogram.types import Message

from .base import Monitor


class EmbyhubMonitor(Monitor):
    name = "EmbyHub"
    chat_name = "emby_hub"
    chat_user = "ednovas"
    chat_keyword = r"注册已开放"
    bot_username = "EdHubot"
    notify_create_name = True

    async def on_trigger(self, message: Message, key, reply):
        await self.client.send_message(self.bot_username, f"/create {self.unique_name}")
        self.log.bind(notify=True).info(f'已向Bot发送用户注册申请: "{self.unique_name}", 请检查结果.')



from pyrogram.types import Message

from .base import Monitor
from .pornemby import PornembyMonitor
from .misty import MistyMonitor
from .bgk import BGKMonitor
from .polo import PoloMonitor
from .viper import ViperMonitor

__ignore__ = True


class TestMonitor:
    class TestReplyMonitor(Monitor):
        name = "回复 测试"
        chat_name = "api_group"
        chat_allow_outgoing = True
        chat_keyword = r"回复\s*(\w+)"
        chat_delay = 1

        def chat_reply(self, message: Message, keys):
            return f'接收到: "{keys[0]}"'

    class TestGroupMonitor(Monitor):
        name = "多组 测试"
        chat_name = "api_group"
        chat_allow_outgoing = True
        chat_keyword = r"多组\s*(\w+)\s+(\w+)"
        chat_delay = 1

        def chat_reply(self, message: Message, keys):
            return f'接收到: "{keys[0]}" => "{keys[1]}"'

    class TestFollowMonitor(Monitor):
        name = "从众 测试"
        chat_name = "api_group"
        chat_allow_outgoing = True
        chat_keyword = r"从众"
        chat_follow_user = 3
        chat_delay = 1
        chat_reply = "我来"

    class TestPornembyExamMonitor(PornembyMonitor.PornembyAnswerMonitor):
        name = "Pornemby 科举 测试"
        chat_name = "api_group"
        chat_user = "embykeeper_test_bot"
        chat_allow_outgoing = True

    class TestPornembyDragonRainMonitor(PornembyMonitor.PornembyDragonRainMonitor):
        name = "Pornemby 红包雨 测试"
        chat_name = "api_group"
        chat_user = "embykeeper_test_bot"
        chat_allow_outgoing = True

    class TestPornembyDoubleMonitor(PornembyMonitor.PornembyDoubleMonitor):
        name = "Pornemby 怪兽自动翻倍 测试"
        chat_name = "api_group"
        chat_user = "embykeeper_test_bot"
        chat_allow_outgoing = True

    class TestMistyMonitor(MistyMonitor):
        name = "Misty 测试"
        chat_name = "api_group"
        chat_allow_outgoing = True
        chat_user = []

    class TestBGKMonitor(BGKMonitor):
        name = "不给看 测试"
        chat_name = "api_group"
        chat_allow_outgoing = True
        chat_user = []

    class TestViperMonitor(ViperMonitor):
        name = "Viper 测试"
        chat_name = "api_group"
        chat_allow_outgoing = True
        chat_user = []

    class TestPoloMonitor(PoloMonitor):
        name = "Polo 测试"
        chat_name = "api_group"
        chat_allow_outgoing = True
        chat_user = []



from datetime import datetime
from ..lock import pornemby_status
from .base import Messager


class PornembyMessager(Messager):
    name = "Pornemby"
    chat_name = "Pornemby"
    default_messages = ["pornemby-common-wl@latest.yaml * 100"]

    async def prepare_send(self, message):
        if pornemby_status.get("nohp", None) and pornemby_status["nohp"] >= datetime.today().date():
            self.log.warning(f"取消发送: 血量已耗尽.")
        else:
            return message



import asyncio
import random
from .base import AnswerBotCheckin

from pyrogram.types import Message


class PornEmbyCheckin(AnswerBotCheckin):
    name = "Pornemby"
    bot_username = "PronembyTGBot2_bot"
    bot_success_pat = r".*?(\d+)$"

    async def start(self):
        if not self.client.me.username:
            self.log.warning(f"签到失败: 需要设置用户名才可生效.")
            return None
        return await super().start()

    async def on_photo(self, message: Message):
        await asyncio.sleep(random.uniform(2, 4))
        async with self.client.catch_reply(self.bot_username) as f:
            try:
                await message.click(0)
            except TimeoutError:
                pass
            try:
                await asyncio.wait_for(f, 10)
            except asyncio.TimeoutError:
                self.log.warning(f"签到失败: 签到无回应, 您可能还没有注册{self.name}.")
                await self.fail()



import asyncio
import random
import time
from typing import Callable, Coroutine, List, Optional, Tuple, Union
import uuid

import tomli
from loguru import logger
from pyrogram import filters
from pyrogram.handlers import MessageHandler
from pyrogram.types import Message
from pyrogram.raw.functions.messages import DeleteHistory

from ..utils import async_partial, truncate_str
from .tele import Client


class Link:
    """云服务类, 用于认证和高级权限任务通讯."""

    bot = "embykeeper_auth_bot"

    def __init__(self, client: Client):
        self.client = client
        self.log = logger.bind(scheme="telelink", username=client.me.name)

    @property
    def instance(self):
        """当前设备识别码."""
        rd = random.Random()
        rd.seed(uuid.getnode())
        return uuid.UUID(int=rd.getrandbits(128))

    async def delete_messages(self, messages: List[Message]):
        """删除一系列消息."""

        async def delete(m: Message):
            try:
                await m.delete(revoke=True)
                text = m.text or m.caption or "图片或其他内容"
                text = truncate_str(text.replace("\n", ""), 30)
                self.log.debug(f"[gray50]删除了 API 消息记录: {text}[/]")
            except asyncio.CancelledError:
                pass

        return await asyncio.gather(*[delete(m) for m in messages])

    async def post(
        self, cmd, condition: Callable = None, timeout: int = 5, retries=3, name: str = None
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        向机器人发送请求.
        参数:
            cmd: 命令字符串
            condition: 布尔或函数, 参数为响应 toml 的字典形式, 决定该响应是否为有效响应.
            timeout: 超时 (s)
            retries: 最大重试次数
            name: 请求名称, 用于用户提示
        """
        for r in range(retries):
            self.log.debug(f"[gray50]禁用提醒 {timeout} 秒: {self.bot}[/]")
            await self.client.mute_chat(self.bot, time.time() + timeout + 5)
            future = asyncio.Future()
            handler = MessageHandler(
                async_partial(self._handler, cmd=cmd, future=future, condition=condition),
                filters.text & filters.bot & filters.user(self.bot),
            )
            await self.client.add_handler(handler, group=1)
            try:
                messages = []
                messages.append(await self.client.send_message(self.bot, f"/start quiet"))
                await asyncio.sleep(0.5)
                messages.append(await self.client.send_message(self.bot, cmd))
                self.log.debug(f"[gray50]-> {cmd}[/]")
                results = await asyncio.wait_for(future, timeout=timeout)
            except asyncio.CancelledError:
                try:
                    await asyncio.wait_for(self.delete_messages(messages), 1.0)
                except asyncio.TimeoutError:
                    pass
                finally:
                    raise
            except asyncio.TimeoutError:
                await self.delete_messages(messages)
                if r + 1 < retries:
                    self.log.info(f"{name}超时 ({r + 1}/{retries}), 将在 3 秒后重试.")
                    await asyncio.sleep(3)
                    continue
                else:
                    self.log.warning(f"{name}超时 ({r + 1}/{retries}).")
                    return None
            else:
                await self.delete_messages(messages)
                status, errmsg = [results.get(p, None) for p in ("status", "errmsg")]
                if status == "error":
                    self.log.warning(f"{name}错误: {errmsg}.")
                    return False
                elif status == "ok":
                    return results
                else:
                    self.log.warning(f"{name}出现未知错误.")
                    return False
            finally:
                await self.client.remove_handler(handler, group=1)

    async def _handler(
        self,
        client: Client,
        message: Message,
        cmd: str,
        future: asyncio.Future,
        condition: Union[bool, Callable[..., Coroutine], Callable] = None,
    ):
        try:
            toml = tomli.loads(message.text)
        except tomli.TOMLDecodeError:
            self.delete_messages([message])
        else:
            try:
                if toml.get("command", None) == cmd:
                    if condition is None:
                        cond = True
                    elif asyncio.iscoroutinefunction(condition):
                        cond = await condition(toml)
                    elif callable(condition):
                        cond = condition(toml)
                    if cond:
                        future.set_result(toml)
                        await asyncio.sleep(0.5)
                        await self.delete_messages([message])
                        return
            except asyncio.CancelledError as e:
                try:
                    await asyncio.wait_for(self.delete_messages([message]), 1)
                except asyncio.TimeoutError:
                    pass
                finally:
                    future.set_exception(e)
            finally:
                message.continue_propagation()

    async def delete_history(self):
        self.log.debug("清空机器人日志记录.")
        await self.client.invoke(DeleteHistory(max_id=0, peer=await self.client.resolve_peer(self.bot)))

    async def auth(self, service: str):
        """向机器人发送授权请求."""
        results = await self.post(f"/auth {service} {self.instance}", name=f"服务 {service.capitalize()} 认证")
        return bool(results)

    async def captcha(self):
        """向机器人发送 Cloudflare 验证码解析请求."""
        results = await self.post(f"/captcha {self.instance}", timeout=240, name="请求跳过验证码")
        if results:
            return [results.get(p, None) for p in ("token", "proxy", "useragent")]
        else:
            return None, None, None

    async def answer(self, question: str):
        """向机器人发送问题回答请求."""
        results = await self.post(f"/answer {self.instance} {question}", timeout=10, name="请求问题回答")
        if results:
            return results.get("answer", None)

    async def send_log(self, message):
        """向机器人发送日志记录请求."""
        results = await self.post(f"/log {self.instance} {message}", name="发送日志到 Telegram")
        return bool(results)



# 该文件用于同机器人 Messager, Monitor 和 Bots 之间的异步锁和通讯

import asyncio

from cachetools import TTLCache

ocrs = TTLCache(maxsize=1024, ttl=3600)
ocrs_lock = asyncio.Lock()

misty_monitors = {}
misty_locks = {}

pornemby_status = {"nohp": None}



import asyncio
import random
from pyrogram.types import Message

from .base import Monitor
from ...utils import async_partial


class JudogMonitor(Monitor):
    name = "剧狗"
    chat_name = "Mulgoreemby"
    chat_keyword = r"剩余可注册人数：\d+"
    bot_username = "mulgorebot"
    notify_create_name = True
    allow_edit = False

    async def init(self):
        channel = await self.client.get_chat("Mulgoreemby")
        self.chat_name = channel.linked_chat.id
        self.log.info(f"已读取剧狗频道关联群: {channel.linked_chat.title}")
        return True

    async def on_trigger(self, message: Message, key, reply):
        wr = async_partial(self.client.wait_reply, self.bot_username)
        msg: Message = await wr("/start")
        if "选择您要使用的功能" in (msg.caption or msg.text):
            await asyncio.sleep(random.uniform(2, 4))
            msg = await wr("🔱账号")
        if "账号管理中心" in (msg.caption or msg.text):
            await asyncio.sleep(random.uniform(2, 4))
            msg = await wr("💡注册")
        if "目前已无可注册资格" in (msg.caption or msg.text):
            return
        else:
            self.log.bind(notify=True).info(f"已在{self.name}抢注成功, 请及时登录查看.")
